<?php

namespace Drupal\Tests\commerce_quickbooks_enterprise\Kernel;

use Drupal\commerce_order\Adjustment;
use Drupal\commerce_order\Entity\Order;
use Drupal\commerce_payment\Entity\Payment;
use Drupal\commerce_payment\Entity\PaymentGateway;
use Drupal\commerce_payment\Entity\PaymentMethod;
use Drupal\commerce_price\Price;
use Drupal\commerce_product\Entity\Product;
use Drupal\commerce_product\Entity\ProductVariation;
use Drupal\commerce_product\Entity\ProductVariationType;
use Drupal\Core\Database\Database;
use Drupal\profile\Entity\Profile;
use Drupal\Tests\commerce\Kernel\CommerceKernelTestBase;

/**
 * Tests exporting data to quickbooks.
 *
 * @requires module commerce_shipping
 * @requires module physical
 *
 * @group commerce_quickbooks_enterprise
 */
class SoapServiceTest extends CommerceKernelTestBase {

  /**
   * {@inheritdoc}
   */
  public static $modules = [
    'migrate_drupal_d8',
    'migrate',
    'commerce_quickbooks_enterprise',
    'entity_reference_revisions',
    'profile',
    'state_machine',
    'commerce_product',
    'commerce_order',
    'commerce_payment',
    'commerce_payment_example',
    'commerce_tax',
    'commerce_shipping',
    'physical',
    'path',
  ];


  /**
   * Testing demo user.
   *
   * @var \Drupal\user\UserInterface
   */
  public $user;

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();

    $this->installEntitySchema('profile');
    $this->installEntitySchema('commerce_order');
    $this->installEntitySchema('commerce_order_item');
    $this->installEntitySchema('commerce_payment');
    $this->installEntitySchema('commerce_payment_method');
    $this->installEntitySchema('commerce_payment_gateway');
    $this->installEntitySchema('commerce_product');
    $this->installEntitySchema('commerce_product_variation');
    $this->installSchema('user', ['users_data']);
    $this->installConfig(self::$modules);

    $user = $this->createUser(['mail' => 'example@example.com', 'pass' => 'abc123'], ['access quickbooks soap service']);
    $this->user = $this->reloadEntity($user);

    // Needed to provide a db connection so we don't fallback to 'migrate'.
    // Otherwise it uses a non-existent connection named 'migrate'.
    $database = Database::getConnectionInfo('default')['default'];
    $config = [
      'target' => 'default',
      'key' => 'default',
      'database' => $database,
    ];
    \Drupal::state()->set('qbe_test_db', $config);
    \Drupal::state()->set('migrate.fallback_state_key', 'qbe_test_db');

    // Create profile data.
    $profile = Profile::create(['uid' => $this->user->id(), 'type' => 'customer']);
    $profile->address = [
      'country_code' => 'AD',
      'locality' => 'Canillo',
      'postal_code' => 'AD500',
      'address_line1' => 'C. Prat de la Creu, 62-64',
      'given_name' => 'John',
      'family_name' => 'Smith',
    ];
    $profile->save();
    $profile = $this->reloadEntity($profile);

    // Turn off title generation to allow explicit values to be used.
    $variation_type = ProductVariationType::load('default');
    $variation_type->setGenerateTitle(FALSE);
    $variation_type->save();

    // Add some products.
    $product = Product::create([
      'type' => 'default',
      'title' => 'Default testing product',
    ]);
    $product->save();
    $variation1 = ProductVariation::create([
      'type' => 'default',
      'sku' => 'SKU1',
      'title' => 'Variation1',
      'status' => 1,
      'price' => new Price('12.00', 'USD'),
    ]);
    $product->addVariation($variation1)->save();
    $variation2 = ProductVariation::create([
      'type' => 'default',
      'sku' => 'SKU2',
      'title' => 'Variation2',
      'status' => 1,
      'price' => new Price('14.00', 'USD'),
    ]);
    $variation2->save();
    $product->addVariation($variation2)->save();

    // Add an order.
    /** @var \Drupal\commerce_order\OrderItemStorageInterface $order_item_storage */
    $order_item_storage = $this->container->get('entity_type.manager')->getStorage('commerce_order_item');
    $order_item1 = $order_item_storage->createFromPurchasableEntity($variation1);
    $order_item1->save();
    $order_item2 = $order_item_storage->createFromPurchasableEntity($variation2);
    $order_item2->save();
    $order = Order::create([
      'type' => 'default',
      'state' => 'completed',
      'billing_profile' => $profile,
      'mail' => $this->user->getEmail(),
      'uid' => $this->user->id(),
      'ip_address' => '127.0.0.1',
      'order_items' => [
        $order_item1,
        $order_item2,
      ],
    ]);
    $order->addAdjustment(new Adjustment([
      'type' => 'tax',
      'label' => 'Tax',
      'amount' => new Price('0.50', 'USD'),
    ]));
    $order->addAdjustment(new Adjustment([
      'type' => 'shipping',
      'label' => 'Shipping',
      'amount' => new Price('5.00', 'USD'),
    ]));
    $order->save();

    // Add a payment.
    $payment_gateway = PaymentGateway::create([
      'id' => 'example',
      'label' => 'Example',
      'plugin' => 'example_onsite',
    ]);
    $payment_gateway->save();
    $payment_gateway = $this->reloadEntity($payment_gateway);
    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $payment_method */
    $payment_method_active = PaymentMethod::create([
      'type' => 'credit_card',
      'card_type' => 'visa',
      'payment_gateway' => 'example',
      'method_id' => 'test_payment',
      // Thu, 16 Jan 2020.
      'expires' => '1579132800',
      'uid' => $user->id(),
    ]);
    $payment_method_active->save();
    $payment_method_active = $this->reloadEntity($payment_method_active);
    // Create an actual payment.
    $payment = Payment::create([
      'payment_gateway' => 'example',
      'payment_method' => $payment_method_active->id(),
      'type' => 'payment_default',
      'remote_id' => '123456',
      'order_id' => $order->id(),
      'amount' => [
        'number' => '31.50',
        'currency_code' => 'USD',
      ],
      'state' => 'completed',
      'test' => TRUE,
    ]);
    $payment->save();
    $payment = $this->reloadEntity($payment);
  }

  /**
   * Tests sales receipt data extraction.
   */
  public function testSalesReceiptExport() {
    /** @var \Drupal\Core\Config\Config $config */
    $config = \Drupal::service('config.factory')->getEditable('commerce_quickbooks_enterprise.quickbooks_admin');
    $exportables = $config->get('exportables');
    $exportables['products'] = FALSE;
    $config->set('exportables', $exportables);
    $config->save();

    $customer_list_id = '80000001-1431947192';
    $order_list_id = '80000001-1431947193';

    /** @var \Drupal\commerce_quickbooks_enterprise\SoapBundle\Services\SoapService $soapService */
    $soapService = \Drupal::service('commerce_quickbooks_enterprise.soap_service');
    $request = new \stdClass();
    $request->ticket = 'UUID';
    $request->strUserName = $this->user->getAccountName();
    $request->strPassword = 'abc123';
    $response = $soapService->authenticate($request);
    $this->assertNotEmpty($response->authenticateResult[0]);
    $this->assertEmpty($response->authenticateResult[1]);

    // Test profile migration.
    $response = $soapService->sendRequestXML($response);
    $result = str_replace(["\n", "\t"], '', $response->sendRequestXMLResult);
    $this->assertEquals('<CustomerAddRq><CustomerAdd><FirstName>John</FirstName><LastName>Smith</LastName><BillAddress><Addr1>C. Prat de la Creu, 62-64</Addr1><City>Canillo</City><PostalCode>AD500</PostalCode><Country>AD</Country></BillAddress><Email>example@example.com</Email></CustomerAdd></CustomerAddRq>', $result);
    $response->response = "<ListID>$customer_list_id</ListID>";
    $response = $soapService->receiveResponseXML($response);
    // We should be only 50% done.
    $this->assertSame($response->receiveResponseXMLResult, 50);
    /** @var \Drupal\migrate\Plugin\Migration $migration */
    $migration = \Drupal::service('plugin.manager.migration')->createInstance('qbe_customer');
    $db_row = $migration->getIdMap()->getRowBySource(['profile_id' => 1]);
    $this->assertEquals($customer_list_id, $db_row['destid1']);

    // Test receipt migration.
    $response = $soapService->sendRequestXML($response);
    $result = str_replace("\n", '', $response->sendRequestXMLResult);
    $date = \Drupal::service('date.formatter')->format(\Drupal::time()->getRequestTime(), 'custom', 'Y-m-d');
    $this->assertEquals("<SalesReceiptAdd><CustomerRef><ListID>80000001-1431947192</ListID><FullName>John Smith</FullName></CustomerRef><RefNumber>1</RefNumber><TxnDate>$date</TxnDate><BillAddress><Addr1>C. Prat de la Creu, 62-64</Addr1><City>Canillo</City><PostalCode>AD500</PostalCode><Country>AD</Country></BillAddress><PaymentMethodRef><FullName>Example</FullName></PaymentMethodRef><SalesReceiptLineAdd><ItemRef><FullName>SKU1</FullName></ItemRef><Desc>Variation1</Desc><Quantity>1</Quantity><Amount>12.00</Amount></SalesReceiptLineAdd><SalesReceiptLineAdd><ItemRef><FullName>SKU2</FullName></ItemRef><Desc>Variation2</Desc><Quantity>1</Quantity><Amount>14.00</Amount></SalesReceiptLineAdd><SalesReceiptLineAdd><ItemRef><FullName>Shipping</FullName></ItemRef><Desc>Shipping fees</Desc><Quantity>1</Quantity><Amount>5.00</Amount></SalesReceiptLineAdd><ItemSalesTaxRef><FullName>Taxes</FullName></ItemSalesTaxRef></SalesReceiptAdd>", $result);
    $response->response = "<TxnID>$order_list_id</TxnID>";
    $response = $soapService->receiveResponseXML($response);
    // We should be 100% done.
    $this->assertSame($response->receiveResponseXMLResult, 100);
    /** @var \Drupal\migrate\Plugin\Migration $migration */
    $migration = \Drupal::service('plugin.manager.migration')->createInstance('qbe_order');
    $db_row = $migration->getIdMap()->getRowBySource(['order_id' => 1]);
    $this->assertEquals($order_list_id, $db_row['destid1']);
  }

  /**
   * Tests invoice data extraction.
   */
  public function testInvoiceExport() {
    /** @var \Drupal\Core\Config\Config $config */
    $config = \Drupal::service('config.factory')->getEditable('commerce_quickbooks_enterprise.quickbooks_admin');
    $exportables = $config->get('exportables');
    $exportables['order_type'] = 'invoices';
    $exportables['products'] = FALSE;
    $config->set('exportables', $exportables);
    $config->save();

    $customerListId = '80000001-1431947192';
    $orderTxnId = '80000001-1431947193';
    $paymentListId = '80000001-1431947194';

    /** @var \Drupal\commerce_quickbooks_enterprise\SoapBundle\Services\SoapService $soapService */
    $soapService = \Drupal::service('commerce_quickbooks_enterprise.soap_service');
    $request = new \stdClass();
    $request->ticket = 'UUID';
    $request->strUserName = $this->user->getAccountName();
    $request->strPassword = 'abc123';
    $response = $soapService->authenticate($request);
    $this->assertNotEmpty($response->authenticateResult[0]);
    $this->assertEmpty($response->authenticateResult[1]);

    // Test profile migration.
    $response = $soapService->sendRequestXML($response);
    $result = str_replace(["\n", "\t"], '', $response->sendRequestXMLResult);
    $this->assertEquals('<CustomerAddRq><CustomerAdd><FirstName>John</FirstName><LastName>Smith</LastName><BillAddress><Addr1>C. Prat de la Creu, 62-64</Addr1><City>Canillo</City><PostalCode>AD500</PostalCode><Country>AD</Country></BillAddress><Email>example@example.com</Email></CustomerAdd></CustomerAddRq>', $result);
    $response->response = "<ListID>$customerListId</ListID>";
    $response = $soapService->receiveResponseXML($response);
    // We should be only 33% done.
    $this->assertSame($response->receiveResponseXMLResult, 33);
    /** @var \Drupal\migrate\Plugin\Migration $migration */
    $migration = \Drupal::service('plugin.manager.migration')->createInstance('qbe_customer');
    $db_row = $migration->getIdMap()->getRowBySource(['profile_id' => 1]);
    $this->assertEquals($customerListId, $db_row['destid1']);

    // Test invoice migration.
    $response = $soapService->sendRequestXML($response);
    $result = str_replace("\n", '', $response->sendRequestXMLResult);
    $date = \Drupal::service('date.formatter')->format(\Drupal::time()->getRequestTime(), 'custom', 'Y-m-d');
    $this->assertEquals("<InvoiceAdd><CustomerRef><ListID>80000001-1431947192</ListID><FullName>John Smith</FullName></CustomerRef><RefNumber>1</RefNumber><TxnDate>$date</TxnDate><BillAddress><Addr1>C. Prat de la Creu, 62-64</Addr1><City>Canillo</City><PostalCode>AD500</PostalCode><Country>AD</Country></BillAddress><PaymentMethodRef><FullName>Example</FullName></PaymentMethodRef><InvoiceLineAdd><ItemRef><FullName>SKU1</FullName></ItemRef><Desc>Variation1</Desc><Quantity>1</Quantity><Amount>12.00</Amount></InvoiceLineAdd><InvoiceLineAdd><ItemRef><FullName>SKU2</FullName></ItemRef><Desc>Variation2</Desc><Quantity>1</Quantity><Amount>14.00</Amount></InvoiceLineAdd><InvoiceLineAdd><ItemRef><FullName>Shipping</FullName></ItemRef><Desc>Shipping fees</Desc><Quantity>1</Quantity><Amount>5.00</Amount></InvoiceLineAdd><ItemSalesTaxRef><FullName>Taxes</FullName></ItemSalesTaxRef></InvoiceAdd>", $result);
    $response->response = "<TxnID>$orderTxnId</TxnID>";
    $response = $soapService->receiveResponseXML($response);
    // We should be 66% done.
    $this->assertSame($response->receiveResponseXMLResult, 66);
    /** @var \Drupal\migrate\Plugin\Migration $migration */
    $migration = \Drupal::service('plugin.manager.migration')->createInstance('qbe_order');
    $db_row = $migration->getIdMap()->getRowBySource(['order_id' => 1]);
    $this->assertEquals($orderTxnId, $db_row['destid1']);

    // Test payment migration.
    $response = $soapService->sendRequestXML($response);
    $result = str_replace(["\n", "\t"], '', $response->sendRequestXMLResult);
    $date = \Drupal::service('date.formatter')->format(\Drupal::time()->getRequestTime(), 'custom', 'Y-m-d');
    $this->assertEquals("<ReceivePaymentAddRq><ReceivePaymentAdd><CustomerRef><ListID>80000001-1431947192</ListID></CustomerRef><TxnDate>$date</TxnDate><RefNumber>123456</RefNumber><PaymentMethodRef><FullName>Example</FullName></PaymentMethodRef><AppliedToTxnAdd><TxnID>80000001-1431947193</TxnID><PaymentAmount>31.50</PaymentAmount></AppliedToTxnAdd></ReceivePaymentAdd></ReceivePaymentAddRq>", $result);
    $response->response = "<ListID>$paymentListId</ListID>";
    $response = $soapService->receiveResponseXML($response);
    // We should be 100% done.
    $this->assertSame($response->receiveResponseXMLResult, 100);
    /** @var \Drupal\migrate\Plugin\Migration $migration */
    $migration = \Drupal::service('plugin.manager.migration')->createInstance('qbe_payment');
    $db_row = $migration->getIdMap()->getRowBySource(['payment_id' => 1]);
    $this->assertEquals($paymentListId, $db_row['destid1']);
  }

  /**
   * Test product export data.
   */
  public function testProductExport() {
    $customerListId = '80000001-1431947192';
    $orderTxnId = '80000001-1431947193';
    $productListId = '80000001-1431947195';
    $productVariationListId = '80000001-1431947196';

    /** @var \Drupal\commerce_quickbooks_enterprise\SoapBundle\Services\SoapService $soapService */
    $soapService = \Drupal::service('commerce_quickbooks_enterprise.soap_service');
    $request = new \stdClass();
    $request->ticket = 'UUID';
    $request->strUserName = $this->user->getAccountName();
    $request->strPassword = 'abc123';
    $response = $soapService->authenticate($request);
    $this->assertNotEmpty($response->authenticateResult[0]);
    $this->assertEmpty($response->authenticateResult[1]);

    // Test profile migration.
    $response = $soapService->sendRequestXML($response);
    $response->response = "<ListID>$customerListId</ListID>";
    $response = $soapService->receiveResponseXML($response);

    // Test product migration.
    $response = $soapService->sendRequestXML($response);
    $result = str_replace(["\n", "\t"], '', $response->sendRequestXMLResult);
    $this->assertEquals("<ItemInventoryAddRq><ItemInventoryAdd><Name>Default testing product</Name></ItemInventoryAdd></ItemInventoryAddRq>", $result);
    $response->response = "<ListID>$productListId</ListID>";
    $response = $soapService->receiveResponseXML($response);
    // We should be 40% done.
    $this->assertSame($response->receiveResponseXMLResult, 40);
    /** @var \Drupal\migrate\Plugin\Migration $migration */
    $migration = \Drupal::service('plugin.manager.migration')->createInstance('qbe_product');
    $db_row = $migration->getIdMap()->getRowBySource(['product_id' => 1, 'langcode' => 'en']);
    $this->assertEquals($productListId, $db_row['destid1']);

    // Test receipt migration.
    $response = $soapService->sendRequestXML($response);
    $response->response = "<TxnID>$orderTxnId</TxnID>";
    $response = $soapService->receiveResponseXML($response);

    // Test product variation migration.
    $response = $soapService->sendRequestXML($response);
    $result = str_replace(["\n", "\t"], '', $response->sendRequestXMLResult);
    $this->assertEquals("<ItemInventoryAddRq><ItemInventoryAdd><Name>Variation1</Name><ParentRef><ListID>$productListId</ListID></ParentRef><SalesPrice>12.00</SalesPrice></ItemInventoryAdd></ItemInventoryAddRq>", $result);
    $response->response = "<ListID>$productVariationListId</ListID>";
    $response = $soapService->receiveResponseXML($response);
    // We should be 80% done.
    $this->assertSame($response->receiveResponseXMLResult, 80);
    /** @var \Drupal\migrate\Plugin\Migration $migration */
    $migration = \Drupal::service('plugin.manager.migration')->createInstance('qbe_product_variation');
    $db_row = $migration->getIdMap()->getRowBySource(['variation_id' => 1, 'langcode' => 'en']);
    $this->assertEquals($productVariationListId, $db_row['destid1']);
    // 2nd product variation.
    $response = $soapService->sendRequestXML($response);
    $result = str_replace(["\n", "\t"], '', $response->sendRequestXMLResult);
    $this->assertEquals("<ItemInventoryAddRq><ItemInventoryAdd><Name>Variation2</Name><ParentRef><ListID>$productListId</ListID></ParentRef><SalesPrice>14.00</SalesPrice></ItemInventoryAdd></ItemInventoryAddRq>", $result);
    $response->response = "<ListID>$productVariationListId</ListID>";
    $response = $soapService->receiveResponseXML($response);
    // We should be 100% done.
    $this->assertSame($response->receiveResponseXMLResult, 100);
    /** @var \Drupal\migrate\Plugin\Migration $migration */
    $migration = \Drupal::service('plugin.manager.migration')->createInstance('qbe_product_variation');
    $db_row = $migration->getIdMap()->getRowBySource(['variation_id' => 2, 'langcode' => 'en']);
    $this->assertEquals($productVariationListId, $db_row['destid1']);
  }

}
